# @cloudbase/functions-framework 使用文档

`@cloudbase/functions-framework` 是一个函数式代码运行框架，本文档主要说明如何使用 `@cloudbase/functions-framework` 框架。

`@cloudbase/functions-framework` 可将如下结构的函数代码：

```js
exports.main = function(event, context) {
  return 'Hello, World!'
}
```

运行起来，并通过 HTTP 请求触发云函数执行：

```sh
curl http://url
# 输出：Hello, World!
```

将函数代码打包后，可以进一步部署在 Docker 等虚拟化环境中。

当前框架支持的 Node.js 版本为 `>=18.0.0`。

## 特性

- 单实例多并发的请求处理模式，更贴近一般 Web 服务的运行方式，支持 SSE、Websocket 等长连接
- 相较于云函数（SCF）提供更高的灵活性，无定制运行时，无需关注函数实例的生命周期
- 原生面向 HTTP，无需理解复杂的自定义事件概念
- 支持加载多函数，支持函数间路由，可直接在不同函数间复用代码
- 提供面向 Web 服务的丰富配置项，如超时时间、并发限制、优雅停止等
- 面向用户的多层次、结构化日志输出，支持自定义日志轮转策略、日志打印内容
- 本地与线上环境同构，支持热重启模式，易于在本地调试

## 安装与运行

执行以下命令安装：

```sh
npm i -g @cloudbase/functions-framework
```

安装后执行以下输出查看是否安装成功：

```sh
tcb-ff -h
```

### 直接通过命令行运行

有如下目录结构：

```sh
/workspace
├── index.js
├── package.json
```

在 `index.js` 中有如下代码：

```js
exports.main = function (event, context) {
  const { httpContext } = context
  const { url, httpMethod } = httpContext
  return `[${httpMethod}][${url}] Hello world!`
}
```

在 `workspace` 目录下执行如下命令，启动监听 `3000` 端口的服务：

```sh
tcb-ff
```

未指定参数时，框架读取的默认函数文件为 `index.js`，默认函数名称为 `main`，默认监听端口为 `3000`。

### 注册函数

有如下目录结构：

```sh
/workspace
├── fn.js
└── package.json
```

在 `fn.js` 中引入`@cloudbase/functions-framework`：

```js
// 引入注册函数的方法
import { event } from '@cloudbase/functions-framework'

// 注册一个命名为 `main` 的函数
event("main", async function (event, context) {
  const { httpContext } = context
  const { url, httpMethod } = httpContext
  return `[${httpMethod}][${url}] Hello world!`
})
```

在 `package.json` 中，使用 `@cloudbase/functions-framework` 作为依赖，并配置启动命令：

```json
{
  "name": "example",
  "version": "1.0.0",
  "type": "module",
  "main": "fn.js",
  "scripts": {
    "start": "npx @cloudbase/functions-framework"
  },
  "dependencies": {
    "@cloudbase/functions-framework": "^0.0.9"
  }
}
```

使用 `npm start` 启动服务。

### 进一步封装

如果希望自行解析参数，可在解析后将解析好的配置作为参数，手动启动服务：

```js
import { startFrameworkService } from '@cloudbase/functions-framework'

// 导出函数供外部调用
export function yourFunction() {
  // 解析参数
  const options = {
    // ...
  }
  // 启动服务
  startFrameworkService(options)
}
```

**如何在封装框架的情况下支持`-w`/`--watch`（热重启）功能？**
由于框架的热重启依赖于启动文件，如果重新封装参数解析，可能导致热重启不起作用。可以参考 `bin/tcb-ff.js` 中的热重启实现，适配自己的 CLI 支持即可。

## 参数说明

可通过 `tcb-ff -h` 打印帮助信息查看所有参数。

| 命令行参数            | 环境变量              | 默认值                              | 示例                                              | 说明                                                                                              |
| --------------------- | --------------------- | ----------------------------------- | ------------------------------------------------- | ------------------------------------------------------------------------------------------------- |
| --port                | PORT                  | 3000                                | --port=3000                                       | 监听的端口                                                                                        |
| --target              | FUNCTION_TARGET       | main                                | --target=hello                                    | 目标函数名称                                                                                      |
| --source              | FUNCTION_SOURCE       | index.js or index.cjs or index.mjs  | --source=main.mjs                                 | 目标函数文件名                                                                                    |
| --signatureType       | SIGNATURE_TYPE        | event                               | --signatureType=event                             | 函数格式，当前仅支持`event`                                                                       |
| --gracefulShutdown    | GRACEFUL_SHUTDOWN     | `NODE_ENV` 不为 `production` 时开启 | --gracefulShutdown=false                          | 是否启用服务的优雅停止                                                                            |
| /                     | KILL_TIMEOUT          | 30000                               | KILL_TIMEOUT=60000                                | 优雅停止时，等待请求处理完成的超时时间                                                            |
| --timeout             | TIMEOUT               | 0                                   | --timeout=5000                                    | Node.js Server 超时时间，见 <https://nodejs.org/api/http.html#servertimeout>                      |
| --keepAliveTimeout    | KEEP_ALIVE_TIMEOUT    | 65000                               | --keepAliveTimeout=10000                          | Node.js Server keep-alive 超时时间，见 <https://nodejs.org/api/http.html#server-keepalivetimeout> |
| --captureStack        | CAPTURE_STACK         | `NODE_ENV` 不为 `production` 时开启 | --captureStack=true                               | 请求出错时，是否返回错误堆栈                                                                      |
| --concurrencyLimit    | CONCURRENCY_LIMIT     | 1000                                | --concurrencyLimit=100                            | 并发请求数限制，0 为不限制                                                                        |
| --noExit              | NO_EXIT               | false                               | --noExit=true                                     | 是否禁用 `process.exit()` 避免进程意外退出                                                        |
| --logHeaderBody       | LOG_HEADER_BODY       | false                               | --logHeaderBody=true                              | 是否在请求日志中记录请求头、请求体、响应头、响应体，如记录，将会限制 body 大小为 2KB              |
| --logEventContext     | LOG_EVENT_CONTEXT     | true                                | --logEventContext=false                           | 是否在请求日志中记录入参 `event` 和 `context`                                                     |
| 载                    |
| --logDirname          | LOG_DIRNAME           | ./logs                              | --logDirname=/var/logs                            | 日志文件目录                                                                                      |
| --maxLogFileSize      | MAX_LOG_FILE_SIZE     | `1024 * 1024 * 128`(128m)           | --maxLogFileSize=`2097152`                        | 单个日志文件最大字节数                                                                            |
| --maxLogFiles         | MAX_LOG_FILES         | 4                                   | --maxLogFiles=5                                   | 最多同时存在的日志文件数量                                                                        |
| -w, --watch           | /                     | false                               | -w                                                | 是否启用热重启模式，如启用，将会在函数代码文件变更时自动重启服务                                  |
| --interceptOutput     | INTERCEPT_OUTPUT      | false                               | --interceptOutput=true                            | 是否拦截`console.log`等标准输出，若拦截，会将标准输出的内容记录到日志文件中                       |
| --extendedContextKey  | EXTENDED_CONTEXT_KEY  | `''`                                | --extendedContextKey=X-Functions-Extended-Context | 指定获取扩展 `context.extendedContext` 的头部字段                                                 |
| --functionsConfigFile | FUNCTIONS_CONFIG_FILE | cloudbase-functions.json            | --functionsConfigFile=myconfig.json               | 多函数路由配置文件名                                                                              |
| --dry-run             | /                     | false                               | --dry-run                                         | 是否启用验证模式，如启用，将不会启动服务，仅检查函数是否可以正常加载                              |

## 调试

有如下目录结构：

```sh
/workspace
├── index.js
├── package.json
```

在 `index.js` 中有如下代码：

```js
exports.main = function (event, context) {
  const { httpContext } = context
  const { url, httpMethod } = httpContext
  return `[${httpMethod}][${url}] Hello world!`
}
```

安装 `@cloudbase/functions-framework` 后，在 `workspace` 目录下执行如下命令，启动调试模式：

```sh
node --inspect $(which tcb-ff)
# 或
node --inspect ./node_modules/.bin/tcb-ff
```

打开调试界面，即可进行调试。

## 构建镜像

### 构建`@cloudbase/functions-framework`镜像

在 `docker/framework` 目录下，提供了构建`@cloudbase/functions-framework`镜像的 Dockerfile，参考其 README 即可构建指定版本的框架镜像。

### 构建函数镜像

使用指定版本的`@cloudbase/functions-framework`镜像构建函数镜像时，可将`docker/Dockerfile`复制到函数代码根目录下，并参考 README 构建任意函数镜像。

## 局限

1. 虽然函数框架提供了许多配置，但不支持自定义 HTTP 服务，也不支持更改提供 HTTP 服务的框架。
2. 函数框架并不和云函数（SCF）兼容，因此无法直接将云函数代码迁移到函数框架中。
3. 函数框架无法做到请求粒度的调度，最小调度单位为函数实例，单个函数实例支持多并发。
4. 由于一个函数实例在逻辑上对应着一个服务而非一个请求，如果划分过多的函数，可能导致服务数量过多。
